// FILin fileechoprocessor
// Copyright (c) 1995-99 Ilya Bogdanov
// See licence.txt for details

#include "general.h"
#include "config.h"
#include "index.h"
#include "dirfile.h"
#include "tic.h"
#include "descript.h"
#include "fidomsg.h"
#include "autocrte.h"

static int tosave;
static int localsubscribe;

void makemsg(
    struct  sMsg *msg )
{
    strcpy( msg->fromname, "FILin" );
    strcpy( msg->toname, "SysOp" );
    msg->isatt = 0;
    msg->iskil = !( mgropt & 4 );    // Leave;
    msg->isdir = mgropt & 8;         // Dir;
    msg->iskfs = 0;
    msg->isrcv = 0;
    msg->ishld = 0;
}

static int isDownlink (struct Sarea *area, struct Slink *rqstor)
{
  int i = 0;
  for (; area->downlink[i] != rqstor && i < num_of_links; i++);
  return i;
}

static int sendfile (struct Slink* rqstor, char* param)
{
  char stemp[FILELEN],temp[TICLINE];
  struct Sarea* area;
  struct find_t f;
  int a = 0;

  if( (area=filelook(param))==NULL  // look in filin.ndx and find on disk
     || (a=_dos_findfirst(filefull(stemp,area->dir,param),_A_NORMAL,&f)))
  {                                 // if not then look thru all areas
    if (a) _dos_findclose(&f);
    area=areas;
    do
      if (!_dos_findfirst(filefull(stemp,area->dir,param),_A_NORMAL,&f))
        goto FOUND;
      else
        _dos_findclose(&f);
    while ((area=area->next)!=NULL);
    return 0;
  }
FOUND:                   // from above
  if ( area->group && strchr(rqstor->groups,area->group)==NULL ) return 0;
  strcpy( param, f.name );          // param could contain wildcards
  filefull( stemp, area->dir, param );
  newtic = 1;
  sprintf( temp, "Area %s\n", area->tag );
  putticline( temp );
  sprintf( temp, "File %s\n", param );
  putticline( temp );
  takedesc( area->dir, param );
  sprintf( temp, "Desc %s\n", in.desc);
  putticline( temp );
  sprintf( temp, "Size %ld\n", f.size );
  putticline( temp );
  sprintf( temp, "Crc %08lX\n", crcvalue(stemp) );
  putticline( temp );
  sprintf( temp, "Seenby %s\n", addrtostr(&la) );
  putticline( temp );
  // requestor should be put in seenby
  sprintf( temp, "Seenby %s\n", addrtostr(&rqstor->addr) );
  putticline( temp );
  seenby[0]=rqstor;     // REsending: requestor is downlink
  outbound( area, rqstor, stemp );
  return 1;
}

void mgrtemplate (struct Slink *rqstor, int templ, char *param, FILE *stream)
{
  struct Sarea *area = areas;
  int i, j, first = 1;
  char sign;
    if( !area ) return;
  if (templ == 'L' || templ == 'Q' || templ == 'P' || templ == 'p')
    do
    {
      sign = ' ';               // default sign
      if (area->group && strchr(rqstor->groups,area->group) == NULL)
         sign = '#';            // restricted area
      if ((i = isDownlink(area,rqstor)) < num_of_links)
      {                         // linked area
        if (templ == 'p')                   // %active :
          if (!area->passive[i]) continue;  // if not passive -- skip
          else
          {
            area->passive[i] = 0;           // else -- activate it
            tosave |= 1;
          }
          sign = '+';
          if (area->passive[i]) sign = '-'; // passive area
          else
             for (j = 0; area->uplink[j] && j < num_of_links; j++)
                if (area->uplink[j] == rqstor)
                {
                  sign = '*';   // receive/send
                  break;
                }
      }                         // unlinked area only in list
      else if (templ == 'Q' || templ == 'P' || templ == 'p') continue;
      if (templ == 'P')                     // %passive :
        if (area->passive[i]) continue;     // if already passive -- skip
        else
        {
          area->passive[i] = 1;             // else -- passivate it
          sign = '-';
          tosave |= 1;
        }                       // if not Hide restricted in list
      if (!(templ == 'L' && sign == '#' && mgropt & 2))
        {
          if (first) first = 0;
          else fputc('\n',stream);
          fprintf(stream,"%c%-20s [%c] %s",sign,area->tag,
                  (area->group)?area->group:' ',area->desc);
        }
    }
    while ((area = area->next) != NULL);
  else if (templ=='S')
  {
    if (!sendfile(rqstor,param))
       logout(stream,'*',"Cannot send file %s",param);
    else
       logout(stream,'*',"File %s was sent to you",param);
  }
  else fputc(templ,stream);              // if not supported macro
}

void service(
    struct Slink    *rqstor,
    char            *rqline,
    FILE            *stream,
    struct sMsg     *rplmsg )
{
    char            command[FILELEN], param[FILELEN];
    FILE            *Fmgr, *fp;
    int             ch;
    struct sMsg     srvmsg;
    Fmgr = fopen(
        strcat( strupr( wordcpy( command, rqline + 1, FILELEN - 4 ) ), ".MGR" )
        , "rt" );
    *( strchr( command, 0 ) - 4 ) = 0;
    if( !Fmgr )
    {
        logout( stream, '*', "Command %%%s: unknown", command );
        return;
    }
    memcpy( &srvmsg, rplmsg, sizeof( struct sMsg ) );
    strcpy( srvmsg.subj, command );
    if( !( fp = fopen( "mgrsrv.fln", "wt" ) ) ) errexit( 1, "mgrsrv.fln" );
    if( word( rqline, 1 ) < strchr( rqline, 0x0D ) )
        wordcpy( param, word( rqline, 1 ), FILELEN );
    else
        *param = 0;
    while( ( ch = fgetc( Fmgr ) ) != EOF )
        if( ch == '@' ) mgrtemplate( rqstor, fgetc( Fmgr ), param, fp );
        else fputc( ch, fp );
    fclose( fp );
    fclose( Fmgr );
    if( msgtext( &srvmsg, "mgrsrv.fln" ) == -1
        || writemsg( outmaildir, &srvmsg ) == -1 )
            logout( stdout, '?', "Command %%%s: failed", command );
    else
    {
        unlink( "mgrsrv.fln" );
        free( srvmsg.text );
        logout( NULL, '%', "Command %%%s: generated", command );
    }
}

struct Sarea *iswildarea(
    char    *pretend )
{
    static struct Sarea *Atemp;
    if( !areas ) return NULL;
    if( !Atemp ) Atemp = areas; // 1st area
    else Atemp = Atemp->next;
    if( Atemp )
        while( wildnicmp( pretend, Atemp->tag, TAGLEN ) && ( Atemp = Atemp->next ) );
    return Atemp;
}

void editauto(          // makes area template passthru
    char    *autoar )
{
    char    *p, template[CONFIGLINE];
    if( !( p = strstr( autoar, "%d" ) ) && !( p = strstr( autoar, "%D" ) ) ) return;
    while( !isspace( *p ) ) --p;    // space before directory
    *p++ = 0;                       // directory beginning
    strcpy( template, autoar );
    while( !isspace( *p ) ) ++p;    // space after directory
    strcat( template, p );
    strcpy( autoar, template );
}    

void rqforward(
    struct  Slink *uplink,
    char    *rqline )
{
    struct  sMsg msg;
    makemsg( &msg );
    strcpy( msg.fromname, "SysOp" );
    strcpy( msg.toname, "AllFix" );
    msg.from = address[0];
    msg.to = uplink->addr;
    strcpy( msg.subj, uplink->pass );
    msg.text = rqline;
    writemsg( outmaildir, &msg );
}    

int forward(
    char    *tag,
    struct  Slink *rqstor )
{
    struct  Slink *Ltemp = links;
    int     group;
    char    rqline[TAGLEN+1] = "+";
    char    template[CONFIGLINE];
    do
    {
        if( Ltemp->arealist )
        {
            if( !strchr( rqstor->groups, *Ltemp->groups )   // restricted group
                || !isinlist( Ltemp->arealist, tag )        // no area in arealist
                || ( group = getgroup( Ltemp ) ) == -1 )    // no autoarea template
                    continue;
            rqforward( Ltemp, strcat( rqline, tag ) );
        // --- autocreating
            strcpy( in.area, tag );
            if( !localsubscribe )
            {
                strcpy( template, autoarea[group].area );
                editauto( autoarea[group].area );           // passthru template
            }    
            else
                --num_of_links;                             // host excluded
            autocreate( Ltemp );
            if( !localsubscribe )
                strcpy( autoarea[group].area, template );   // original template
            else
                ++num_of_links;                             // host is back
            return 1;
        }
    }
    while( Ltemp = Ltemp->next );
    return 0;
}

void connect(
    struct  Sarea *area,
    struct  Slink *rqstor,
    FILE    *stream )    
{
    int     i;
    if( isDownlink( area, rqstor ) < num_of_links )
        logout( stream, '*', "You are already connected to %s", area->tag );
    else if( area->group && strchr( rqstor->groups, area->group ) == NULL )
        logout( stream, '*', "Area %s is restricted for you", area->tag );
    else
    {
        if( addrcmp( &rqstor->addr, &address[0] ) )   // local subscriber
            area->downlink[num_of_links-1] = rqstor;
        else    
        {
            for( i = 0; area->downlink[i]; i++ );   // there SHOULD be NULL element
            area->downlink[i] = rqstor;
        }
    // --- bidirect connection
        if( rqstor->bidirect && area->bidirect )
        {
        // --- search NULL or rqstor uplink element
            for( i = 0; area->uplink[i] && area->uplink[i] != rqstor; i++ );
                area->uplink[i] = rqstor;
        // --- eliminate extra rqstor uplink
            for( ++i ; area->uplink[i] != rqstor && i < num_of_links; i++ );                    
            if( i < num_of_links ) area->uplink[i] = NULL;
        }    
        logout( stream, '+', "Connected to area %s", area->tag );
        tosave |= 1;
    }
}

void subscribe(
    struct  Slink *rqstor,
    char    *rqline,
    FILE    *stream )
{
    char    tag[TAGLEN];
    struct  Sarea *area;
    int     i, found;
// --- get areatag
    strupr( wordcpy( tag, rqline + ( *rqline == '-' || *rqline == '+' ) , TAGLEN ) );
// --- start matching areas
    if( ( area = iswildarea( tag ) ) == NULL )
    {
        if( !rqstor->forward || !forward( tag, rqstor ) )
        {
            logout( stream, '*', "Area %s doesn't exist", tag );
            return;
        }
        else
        {
            logout( stream, '*', "Request for area %s forwarded", tag );
            area = iswildarea( tag );
        }
    }
// --- unsubscribe
    if( *rqline == '-' )
        do
        {
            found = 0;
        // --- look through all area links
            for( i = 0; i < num_of_links; i++ )
            {
                if( area->downlink[i] == rqstor )
                {
                    area->downlink[i] = NULL;
                    found = 1;
                }
                if( area->uplink[i] == rqstor )
                {
                    area->uplink[i] = NULL;
                    found = 1;
                }
            }
            if( found )
            {
                logout( stream, '-', "Disconnected from area %s", area->tag );
                tosave |= 1;
            }
            else
                logout( stream, '*', "You are not connected to area %s", area->tag );
        }
        while( area = iswildarea( tag ) );
// --- subscribe
    else
        do connect( area, rqstor, stream );
        while( area = iswildarea( tag ) );
}

void rq2rpl(
    struct sMsg *rplmsg,
    struct sMsg *rqmsg )
{
    makemsg( rplmsg );
    rplmsg->from = rqmsg->to;
    rplmsg->to = rqmsg->from;
    strcpy( rplmsg->subj, "Answer from FILin areamanager" );
}

int isrq(
    char        *rqpath,
    struct sMsg *rqmsg )
{
    if( readmsg( rqpath, rqmsg, address[0].zone ) == -1 )
    {
        logout( stdout, '?', "Cannot read %s", rqpath );
        return 0;
    }
    if( isaddress( &rqmsg->to ) == -1
        || !strstr( aliases, strupr( rqmsg->toname ) )
        || rqmsg->isrcv )
    {
        free( rqmsg->text );
        return 0;
    }
    return 1;
}

void msgrcvd(
    char    *rqpath )
{
    FILE    *fp;
    int     c;
    if( mgropt & 1 ) unlink( rqpath );
    else if( fp = fopen( rqpath, "r+b" ) )
    {
        fseek( fp, 186L, SEEK_SET );
        c = fgetc( fp ) | 4;
        fseek( fp, 186L, SEEK_SET );
        fputc( c, fp );
        fclose( fp );
    }
}

static void areas2local(           // <- areamgr
    struct Slink *rqstor )    
{
    struct Sarea *area = areas;
    if( !area ) return;
    do
        if( !ISPASSTHRU( area ) )
            area->downlink[num_of_links] = rqstor;
    while( area = area->next );
    ++num_of_links;
}

static void local2areas(           // <- areamgr
    struct Slink *rqstor )    
{
    struct Sarea *area = areas;
    char templine[CONFIGLINE], *w;
    int j;
    --num_of_links;
    if( !area ) return;
    do
    {
        if( area->downlink[num_of_links] == rqstor )    // host connected
        {
            if( ISPASSTHRU( area ) )
            {
                tpl2str(                                // -> autocreate 
                    autoarea[ getgroup( area->uplink[0] ) ].area,
                    templine,
                    area->tag );
                for( j = 2; *( w = word( templine, j ) ) && ISPASSTHRU( area ); j++ )
                    switch( *w )
                    {
                        case '+': case '$': case '^': case '"':
                            break;
                        default:
                            wordcpy( area->dir, w, DIRLEN );
                    } 
            }
        }
        else                                            // host disconnected
        {
            if( !ISPASSTHRU( area ) )
                strcpy( area->dir, "" );
        }
    }
    while( area = area->next );
}

void areamgr(
    char    *rqpath )
{
    struct  sMsg rqmsg, rplmsg;
    struct  Slink *rqstor;
    char    *Pbuf;
    FILE    *fp;
// --- load and check rqmsg
    if( !isrq( rqpath, &rqmsg ) ) return;
    logout( stdout, 0, "  %s from %s", strrchr( rqpath, '\\' ) + 1, addrtostr( &rqmsg.from ) );
    rq2rpl( &rplmsg, &rqmsg );
    localsubscribe = 0;
    if( addrcmp( &rqmsg.from, &address[0] ) )   // local subscriber
    {
        if( stricmp( mgrpass, rqmsg.subj ) )
        {
            logout( stdout, '?', rplmsg.text = "Invalid mgr password" );
            writemsg( outmaildir, &rplmsg );
            msgrcvd( rqpath );
            return;
        }
        localsubscribe = 1;
        rqstor = NEW( struct Slink );
        rqstor->addr = address[0];
        rqstor->forward = 1;
        strcpy( rqstor->groups, "ABCDEFGHIJKLMNOPQRSTUVWXYZ" );
        areas2local( rqstor );
    }
    else if( !( rqstor = islink( &rqmsg.from ) ) || stricmp( rqstor->pass, rqmsg.subj ) )
    {
        logout( stdout, '?', rplmsg.text = "Invalid link or password" );
        if( writemsg( outmaildir, &rplmsg ) == -1 ) logout( stdout, '?', "Answer failed" );
        free( rqmsg.text );
        msgrcvd( rqpath );
        return;
    }
    tosave = 0;
// --- process request    
    Pbuf = rqmsg.text;
    if( !( fp = fopen( "mgranswr.fln", "wt" ) ) ) errexit( 1, "mgranswr.fln" );
    while( *Pbuf )
    {
        if( *Pbuf != 0x0D )
        {
            while( isspace( *Pbuf ) ) ++Pbuf;
            if( *Pbuf == '-' || *Pbuf == '+' || isalnum( *Pbuf ) || *Pbuf == '?' || *Pbuf == '*' )
                subscribe( rqstor, Pbuf, fp );
            else if( *Pbuf == '%' )
                service( rqstor, Pbuf, fp, &rplmsg );
            if( !( Pbuf = strchr( ++Pbuf, 0x0D ) ) ) break;
        }
        if( *++Pbuf == 0x0A ) ++Pbuf;
    }
    fclose( fp );
// --- post answer    
    if( msgtext( &rplmsg, "mgranswr.fln" ) == -1 )
    {
        logout( stdout, '?', "Answer failed" );
        return;
    }
    unlink( "mgranswr.fln" );
    if( *rplmsg.text != 0 ) writemsg( outmaildir, &rplmsg );
    free( rqmsg.text );
    msgrcvd( rqpath );
// --- rewrite areas.ctl
    if( localsubscribe ) local2areas( rqstor );
    if( tosave ) save_areas();
}

void delarea(
    struct Sarea *area )
{
    struct Sarea *prevarea = areas;
    if( area == areas )
        areas = area->next;             // delete 1st area
    else
    {
        while( prevarea->next != area )
            prevarea = prevarea->next;  // prevarea cannot be NULL
        prevarea->next = area->next;
    }
    logout( stdout, 0, "  Orphaned area %s killed", area->tag );
}

void killorphans( void )
{
    struct  Sarea *area = areas;
    int     i, changed = 0;
    char    rqline[TAGLEN+1];
    if( !areas ) return;
    do
    {
        if( ISPASSTHRU( area ) )
        {
            for( i = 0; i < num_of_links && !area->downlink[i]; ++i );
            if( i == num_of_links )             // no downlinks
            {
                for( i = 0; i < num_of_links && !area->uplink[i]; ++i );
                if( i < num_of_links )         
                {
                    rqforward(
                        area->uplink[i],
                        strcat( strcpy( rqline, "-" ), area->tag ) );
                }
                delarea( area );
                changed = 1;
            }
        }
    }
    while( area = area->next );
    if( changed ) save_areas();
}

