// FILin fileechoprocessor
// Copyright (c) 1995-99 Ilya Bogdanov
// See licence.txt for details

#include "general.h"
#include "dirfile.h"
#include "autocrte.h"

struct Sarea *parsearea(
    char                *l )
{
    static struct Sarea *Atemp;
    int                 j, k;
    char                *w, tempaddr[ADDRLEN];
    struct sAddr        temp;
    struct Sarea        *old = Atemp;
    Atemp = ( struct Sarea * ) calloc( 1, sizeof( struct Sarea ) );
    if( !areas ) areas = Atemp;
    else old->next = Atemp;
    wordcpy( Atemp->tag, word( l, 1 ), TAGLEN );
    if( isarea( Atemp->tag ) != Atemp ) errexit( 4, Atemp->tag );
    for( j = 2; *( w = word( l, j ) ); j++ )
        switch( *w )
        {
            case '+':
                switch( *( w + 1 ) )
                {
                    case 'R': case 'r': Atemp->replace = 1; break;
                    case 'M': case 'm': Atemp->magic = 1; break;
                    case 'B': case 'b': Atemp->bidirect = 1; break;
                }
                break;
            case '$':
                if( isupper( *( w + 1 ) ) ) Atemp->group = *( w + 1 );
                else errexit( 8, l );   // incorrect group
                break;
            case '^':
                wordcpy( tempaddr, w + 1, ADDRLEN );
                if( parseaddr( &temp, tempaddr, &la ) )
                {
                    for( k = 0; k < 10 && !addrcmp( &temp, &address[k] ); k++ );
                    if( k < 10 )
                    {
                        Atemp->aka = k;
                        break;
                    }
                }
                errexit( 6, tempaddr ); // invalid station address
            case '"':
                for( k = 0; k < DESCLEN - 1 && w[ k + 1 ] != '"'; k++ )
                    Atemp->desc[ k ] = w[ k + 1 ];
                goto FINISH;            // description is the last
                break;
            default:
                wordcpy( Atemp->dir, w, DIRLEN );
        } 
FINISH:
    if( !ISPASSTHRU( Atemp ) )
        if( !makexist( Atemp->dir ) )
            errexit( 2, Atemp->dir );
    Atemp->uplink = ( struct Slink** )
        calloc( num_of_links, sizeof( struct Slink* ) );
    Atemp->downlink = ( struct Slink** )
        calloc( num_of_links + 1, sizeof( struct Slink* ) );    // + 1 for locallink (areamgr)
    Atemp->passive = ( int * ) calloc( num_of_links + 1, sizeof( int ) );
    return Atemp;
}

void parseupdown (char* l, struct Sarea* Atemp, int down)
{
  char *link, tempaddr[ADDRLEN];
  int j2, j1, j = 0;
  struct sAddr templink;
  struct Slink* Ltemp;

    if (!Atemp || !l) return;
  memcpy(&la,&address[Atemp->aka],SADDR);
  for (j1 = 1; *(link = word(l,j1)); j1++)
  {
    if (*link == '-')
      wordcpy(tempaddr,link+1,ADDRLEN);
    else
      wordcpy(tempaddr,link,ADDRLEN);
    if (!parseaddr(&templink,tempaddr,&la) || (Ltemp = islink(&templink)) == NULL)
      errexit(7,tempaddr);
//    {
//      logout(stdout,'?',"Undefined link '%s' (Area %s)",tempaddr,Atemp->tag);
//      exit(0);
//    }
    if (!down)
    {
      for (j2 = 0; j2 < j && Atemp->uplink[j2] != Ltemp; j2++);
      if (j2 == j) Atemp->uplink[j++] = Ltemp;
    }
    else
    {
      for (j2 = 0; j2 < j && Atemp->downlink[j2] != Ltemp; j2++);
      if (j2 == j)
      {
        if (*link == '-') Atemp->passive[j] = 1;
        Atemp->downlink[j++] = Ltemp;
      }
    }
  }
}

void load_areas (void)
{
  FILE* config;
  char line[CONFIGLINE];
  unsigned char* l = line;
  struct Sarea* Atemp = NULL;
  int i = 0;

  if ((config = fopen("areas.ctl","rt")) == NULL) errexit(1,"areas.ctl");
  printf("\n  areas.ctl");
//  logout(stdout,0," ");

  while (fgets(line, CONFIGLINE, config) != NULL)
  {
    i++;
    while (isspace(*l)) l++;
    if (*l != ';' && *l)
    {
      if (!strnicmp(l,"Ar",2)) Atemp = parsearea(l);
      else if (!strnicmp(l,"Re",2)) parseupdown(l,Atemp,0);
      else if (!strnicmp(l,"Se",2)) parseupdown(l,Atemp,1);
      else logout(stdout,'?',"Unknown keyword in line %d",i);
    }
    l = line;
  }
  fclose(config);
}

void write_area (FILE* config1, struct Sarea* Tarea)
{
  char tempaddr[ADDRLEN];
  int i;

  fprintf(config1,"Area %-12s %s %s %s %c%c %c%s %s %c%s%c",
                        Tarea->tag,
                        (Tarea->replace)?"+R":"  ",
                        (Tarea->magic)?"+M":"  ",
                        (Tarea->bidirect)?"+B":"  ",
                        (Tarea->group)?'$':' ',
                        (Tarea->group)?Tarea->group:' ',
                        (Tarea->aka)?'^':' ',
                        (Tarea->aka)?addrtostr(&address[Tarea->aka]):"",
                        Tarea->dir,
                        (*Tarea->desc)?'"':' ',
                        Tarea->desc,
                        (*Tarea->desc)?'"':' ');
  fprintf(config1,"\nReceiveFrom");
  for (i = 0; i < num_of_links; i++)
    if (Tarea->uplink[i] != NULL)
      fprintf(config1," %s",shrtaddr(tempaddr,&Tarea->uplink[i]->addr,&address[Tarea->aka]));
  fprintf(config1,"\nSendTo");
  for (i = 0; i < num_of_links; i++)
    if (Tarea->downlink[i] != NULL)
      fprintf(config1," %s%s",(Tarea->passive[i])?"-":"",
              shrtaddr(tempaddr,&Tarea->downlink[i]->addr,&address[Tarea->aka]));
  fprintf(config1,"\n");
}

void save_areas (void)
{
  FILE* config,* config1;
  char line[CONFIGLINE];
  char* l = line;
  struct Sarea* Tarea = areas;

  if ((config  = fopen("areas.ctl","rt")) == NULL) errexit(1,"areas.ctl");
  if ((config1 = fopen("areas.$$$","wt")) == NULL) errexit(1,"areas.$$$");
  logout(stdout,'*',"Rewriting areas.ctl");

  while (fgets(line, CONFIGLINE, config) != NULL)
  {
    while (isspace(*l)) l++;
    if (!strnicmp(l,"Area",4))
    {
      if (Tarea != NULL)
      {
        write_area(config1,Tarea);
        Tarea = Tarea->next;
      }
    }
    else if (strnicmp(l,"Rece",4) && strnicmp(l,"Send",4))
      fprintf(config1,"%s",line);
    l = line;
  }
  while (Tarea)
  {
    fputc('\n',config1);
    write_area(config1,Tarea);
    Tarea = Tarea->next;
  }

  fclose(config1);
  fclose(config);
  unlink("areas.old");
  if( rename( "areas.ctl", "areas.old" ) || rename( "areas.$$$", "areas.ctl" ) )
    errexit( 0, "Cannot rename areas configuration file" );
}

void load_config (void)
{
  FILE *config;
  char line[CONFIGLINE],tempaddr[ADDRLEN],logfile[PATHLEN] = "";
  unsigned char *l = line;
  int k,j,i = 0,x = 0;
  struct Slink* Ltemp;
  struct sAddr temp;

  if ((config = fopen("filin.ctl","rt")) == NULL) errexit(1,"filin.ctl");
  printf("\n  filin.ctl");
  while (fgets(line, CONFIGLINE, config) != NULL)
  {
    x++;                                // line number
    while (isspace(*l)) l++;
    if (*l != ';' && *l)
    {
      if (!strnicmp(l,"Addr",4))
      {
        k = 0;                          // AKA number
        for (j = 1; *wordcpy(tempaddr,word(l,j),ADDRLEN) && k < AKA; j++)
          if (!parseaddr(&address[k++],tempaddr,&la)) errexit(6,tempaddr);
        memcpy(&la,&address[0],SADDR);
      }
      else if (!strnicmp(l,"FileD",5))
      {
        k = 0;                          // FileDir number
        for (j = 1; *word(l,j) && k < FILEDIR; j++)
          wordcpy(inbound[k++] = calloc(1,DIRLEN),word(l,j),DIRLEN);
        strcpy(filedir,inbound[0]);
      }
      else if (!strnicmp(l,"Tics",4))
      {
        if (!makexist(wordcpy(ticsdir,word(l,1),DIRLEN))) errexit(2,ticsdir);
        if (*word(l,2)) ticsdays = atoi(word(l,2));
      }
      else if (!strnicmp(l,"Lo",2)) {
        wordcpy(logfile,word(l,1),PATHLEN);
        if (toupper(*word(l,2)) == 'L') loglevel = 1;
      }
      else if (!strnicmp(l,"NewA",4))
      {
        wordcpy(newarea,word(l,1),PATHLEN);
        if (*word(l,2) == '+') postsw |= 1;
      }
      else if (!strnicmp(l,"NewF",4))
      {
        wordcpy(newfile,word(l,1),PATHLEN);
        if (*word(l,2) == '+') postsw |= 2;
        wordcpy(postnfr,word(l,3),PATHLEN);
      }
      else if (!strnicmp(l,"Po",2)) 
        wordcpy(postdir,word(l,1),DIRLEN);
      else if (!strnicmp(l,"Alia",4))
#ifndef _UNIX_
        strupr(
#endif
            strcat(aliases,word(l,1))
#ifndef _UNIX_
            )
#endif
            ;
      else if (!strnicmp(l,"AreaM",5)) {
        wordcpy(inmaildir,word(l,1),DIRLEN);
        if (!makexist(inmaildir)) errexit(2,inmaildir);
        if (*word(l,2)) {
          wordcpy(outmaildir,word(l,2),DIRLEN);
          if (!makexist(outmaildir)) errexit(2,outmaildir);
        }
        else strcpy(outmaildir,inmaildir);
      }
      else if (!strnicmp(l,"Mgr",3))
        for (j = 1; *word(l,j); j++)
        {
            switch( *word( l, j ) )
            {
                case 'K': case 'k': mgropt |= 1; break;     // Kill
                case 'H': case 'h': mgropt |= 2; break;     // Hide
                case 'L': case 'l': mgropt |= 4; break;     // Leave
                case 'D': case 'd': mgropt |= 8; break;     // Dir
                case '#': wordcpy( mgrpass, word( l, j ) + 1, PASSLEN );
            }
        }
      else if (!strnicmp(l,"Auto",4)) 
        loadauto(l);
      else if (!strnicmp(l,"FreqA",5)) wordcpy(aliasfile,word(l,1),PATHLEN);
      else if (!strnicmp(l,"DLC",3))
      {
        wordcpy(dlc,word(l,1),DLC-1);
        strcat(dlc," ");
      }
      else if (!strnicmp(l,"Dupe",4))
         switch (*word(l,1))
            {
            case 'M': case 'm': dupeopt = MOVE; break;
            case 'K': case 'k': dupeopt = KILL; break;
            case 'F': case 'f': dupeopt = FORCE;
            }
      else if (!strnicmp(l,"Outb",4))
      {
        if      (!strnicmp(word(l,1),"Box",3)) outtype = BOX;
        else if (!strnicmp(word(l,1),"Att",3)) outtype = AMA;
        else if (!strnicmp(word(l,1),"Bin",3)) outtype = BSO;
        wordcpy(outdir,word(l,2),DIRLEN);
        if (!makexist(outdir)) errexit(2,outdir);
        if( *word( l, 3 ) && *word( l, 3 ) != '+' ) wordcpy( outb, word( l, 3 ), DIRLEN );
        if( !strnicmp( word( l, 3 ), "+S", 2 ) || !strnicmp( word( l, 4 ), "+S", 2 ) )
            outshort = 1;
        if( !makexist( outb ) ) errexit( 2, outb );
      }
      else if (!strnicmp(l,"Link",4))
      {
        if (!parseaddr(&temp,wordcpy(tempaddr,word(l,1),ADDRLEN),&la))
            errexit(7,tempaddr);
        if (islink(&temp)) errexit(4,addrtostr(&temp));
        Ltemp = links;
        if ((links = (struct Slink *) calloc(1,sizeof(struct Slink))) == NULL)
            errexit(3,"links");
        links->addr = temp;
        links->autocr = false;
        links->hold = false;
        links->crc = 1;
        for (j = 2; *word(l,j); j++)
            switch (*word(l,j))
            {
                case '#':
                    wordcpy(links->pass,word(l,j)+1,PASSLEN);
                break;
                case '$': 
                    wordcpy(links->groups,word(l,j)+1,GROUPSLEN);
                    for (k = 0; links->groups[k]; k++)
                        if (!isupper(links->groups[k])) errexit(8,l);
                break;
                case '+':
                    switch (toupper(*(word(l,j)+1)))
                    {
                        case 'A': links->autocr = true; break;
                        case 'H': links->hold = true; break;
                        case 'B': links->bidirect = 1; break;
                        case 'F': links->forward = 1; break;
                    }
                break;
                case '-':
                    if (toupper(*(word(l,j)+1)) == 'C') links->crc = 0;
                break;
                case '%':
                    links->days = atoi(word(l,j)+1);
                break;
                case '@':
                    links->arealist = ( char * ) malloc( PATHLEN );
                    wordcpy( links->arealist, word( l, j ) + 1, PATHLEN );
                break;
            }
        i++;
        links->next = Ltemp;
      }
      else logout(stdout,'?',"Unknown keyword in line %d",x);
    }
    l = line;
  }
  fclose(config);
  num_of_links = i;
  seenby = (struct Slink**) calloc(num_of_links+1,sizeof(struct Slink*));
  if (outtype && !makexist(outb)) errexit(2,outb);
  if (links == NULL) errexit( 0, "No links defined" );
  if( !autoarea[0].area ) errexit( 0, "Default AutoArea not defined" );
  if (*logfile)
    if ((log = fopen(logfile,"at")) == NULL) errexit(1,logfile);
    else fprintf(log,"\n");
}

void loadhat (char* hat)
{
#define LINE 256
#define SECTION 11
  FILE* Fhat;
  char line[LINE], temppath[PATHLEN], section[SECTION];
  unsigned char* l = line;
  int load = 0, is = 0, i = 0;

  if ((Fhat = fopen("hatch.ctl","rt")) == NULL) errexit(1,"hatch.ctl");
  logout(stdout,0,"  hatch.ctl (section %s)",hat);
  while (fgets(line, LINE, Fhat) != NULL)
  {
    i++;
    while (isspace(*l)) l++;
    if (*l != ';' && *l)
    {
      if (*l == 'S' || *l == 's')
      {
        if (load)
        {
          hatch(temppath);
          free(in.desc);
          memset(&in,0,sizeof(in));
          load = 0;
        }
        wordcpy(section,word(l,1),SECTION);
        if (!strnicmp(section,hat,SECTION))
        {
            load = 1;
            if (!is) is = 1;
        }
      }
      else if (load)
      {
        if (*l == 'A' || *l == 'a') wordcpy(in.area,word(l,1),TAGLEN);
        else if (*l == 'F' || *l == 'f') wordcpy(temppath,word(l,1),PATHLEN);
        else if (*l == 'D'|| *l == 'd')
        {
          in.desc = (char*)calloc(strlen(l)-4,1);
          strcpy(in.desc,word(l,1));
          LASTCHAR(in.desc) = 0;
        }
        else if (*l == 'M' || *l == 'm') wordcpy(in.magic,word(l,1),MAGICLEN);
        else if (*l == 'R' || *l == 'r') wordcpy(in.repl,word(l,1),FILELEN);
        else logout(stdout,'?',"Unknown keyword in line %d",i);
      }
    }
    l = line;
  }
  if (load) hatch(temppath);
  fclose(Fhat);
  if (!is) errexit(0,"Section not found");
}
