#include "posixdir.h"

static void free_struct( struct find_t *ft )
{
	if( ft->mask )
		free( ft->mask );
	if( ft->path )
		free( ft->path );
	if( ft->dirp )
		close( ft->dirp );
}

static void init_struct( struct find_t *ft )
{
	ft->mask = ft->path = NULL;
	ft->dirp = NULL;
	ft->direntp = NULL;
	ft->attr = 0;
}

static int search_file( struct find_t *ft )
{
	ft->direntp = readdir( ft->dirp );
	while( ft->direntp != NULL )
	{
		if( !fnmatch( ft->mask, ft->direntp->d_name, 0 ))
		{
			if( ft->direntp->d_stat.st_mode & ft->attr )
			{
				strcpy( ft->name, ft->direntp->d_name );
				ft->size = ft->direntp->d_stat.st_size;
				return 0;
			}
		}
		ft->direntp = readdir( ft->dirp );
	}

	return 1;
}

static char *make_mask( char *old_mask )
{
	static char unix_mask[_MAX_PATH*4];
	char *char_format = "[%c%c]";
	int i=0;
	
	memset( unix_mask, '\0', sizeof( unix_mask ));
	for( ; i<strlen(old_mask); i++ )
	{
		if( isalpha(old_mask[i]) )
		{
			sprintf( unix_mask+strlen( unix_mask ), char_format,
				tolower( old_mask[i] ), toupper( old_mask[i] ));
		}else
		{
			unix_mask[strlen(unix_mask)] = old_mask[i];
		}
	}

	return unix_mask;
}

int _dos_findfirst( char *path_mask, int attr, struct find_t *ft )
{
	char *ptr;
	char c;

	init_struct( ft );

	/*
	 *  splitting path & mask
	 */
	ptr = (char*)strrchr( (const char*)path_mask, '/' );
	if( ptr )
	{
		ft->mask = (char*)strdup( make_mask( ptr+1 ));
		c = *ptr;
		*ptr = 0;
		ft->path = (char*)strdup( path_mask );
		*ptr = c;
	}else
	{
		ft->path = (char*)strdup( "." );
		ft->mask = (char*)strdup( make_mask( path_mask ));
	}

	ft->dirp = opendir( ft->path );
	if( ft->dirp == NULL )
	{
		free_struct( ft );
		return 1;
	}
	ft->attr = attr;

	return search_file( ft );
}

int _dos_findnext( struct find_t *ft )
{
	return search_file( ft );
}


int _dos_findclose( struct find_t *ft )
{
	free_struct( ft );

	return 0;
}
